/*
 * Buffer addition routines.
 *
 * Copyright 1998 by Gray Watson
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted at http://256.com/gray/
 *
 * $Id: addbuf.h,v 1.7 2000/03/07 21:12:38 gray Exp $
 */

#ifndef __ADDBUF_H__
#define __ADDBUF_H__

#include <stdarg.h>				/* for ... */

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * int addbuf_put
 *
 * DESCRIPTION:
 *
 * Add to a buffer a list of (const void *data, const int data_size)
 * pairs, advancing the buffer pointer so it is always at the end of
 * the data.  The list of items are terminated if the data pointer is
 * 0L.  If the length is < 0 it will add a string _including_ the '\0'
 * to the buffer.  Data items of length 0 are ignored.
 *
 * NOTE: this routine is exactly the same as addbuf_put except the \0 is
 * included if the length is < 0.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to a buffer pointer.  We will be adding the data
 * items to this location and then advancing the pointer past the new
 * data.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will add
 * the data items to the buffer until it reaches this pointer.  If it
 * goes to write to this address it will return -1.
 *
 * ... - Variable arguments list of (const void *data, const int
 * data_size) pairs.  If the data_size is < 0 it will add a string
 * _including_ the '\0' to the buffer.
 */
extern
int	addbuf_put(void **buf_pp, const void *bounds_p, ...);

/*
 * int addbuf_put_str
 *
 * DESCRIPTION:
 *
 * Add to a buffer a list of (const char *string, const int string_len)
 * pairs, advancing the buffer pointer so it is always at the end of
 * the data.  The list of items are terminated if the string pointer
 * is 0L.  If the string_len is < 0 it will add a string _not_
 * including the '\0' to the buffer.  Strings of length 0 are ignored.
 *
 * NOTE: this routine is exactly the same as addbuf_put except the \0 is
 * not included if the length is < 0.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to a buffer pointer.  We will be adding the data
 * items to this location and then advancing the pointer past the new
 * data.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will add
 * the data items to the buffer until it reaches this pointer.  If it
 * goes to write to this address it will return -1.
 *
 * ... - Variable arguments list of (const char *string, const int
 * string_len) pairs.  If the string_len is < 0 it will add a string
 * _not_ including the '\0' to the buffer.  Strings of length 0 are
 * ignored.
 */
extern
int	addbuf_put_str(void **buf_pp, const void *bounds_p, ...);

#ifndef NO_VSNPRINTF
/*
 * int addbuf_snprintf
 *
 * DESCRIPTION:
 *
 * Add to a buffer a printf-like format and a variable number of
 * arguments.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to a buffer pointer.  We will be adding the data
 * items to this location and then advancing the pointer past the new
 * data.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will add
 * the data items to the buffer until it reaches this pointer.  If it
 * goes to write to this address it will return -1.
 *
 * format - Printf-style format.
 *
 * ... - Arguments associated to the format.
 */
extern
int	addbuf_snprintf(void **buf_pp, const void *bounds_p,
			const char *format, ...);
#endif /* ifndef NO_VSNPRINTF */

/*
 * int addbuf_get
 *
 * DESCRIPTION:
 *
 * Get from a buffer a data which is unpacked into (void *data, const int
 * data_size) pairs, advancing the buffer pointer so it is always at
 * the end of the data.  The list of items are terminated if the data
 * pointer is 0L.  If the length is < 0 it will get a string including
 * its 0L to the buffer.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to our buffer which we will be adding information to
 * and advancing.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will get
 * the data items from the buffer until it reaches this pointer.  If
 * it goes to access this pointer, it will return -1.
 *
 * ... - Variable arguments list of (void *data, const int data_size)
 * pairs.
 */
extern
int	addbuf_get(void **buf_pp, const void *bounds_p, ...);

/*
 * int addbuf_put_uint
 *
 * DESCRIPTION:
 *
 * Add to a buffer a packed decimal value, advancing the buffer pointer
 * so it is always at the end of the data.  It will pack the decimal
 * into as few bytes as possible so addbuf_get_uint must be used on
 * the other side to unpack.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to our buffer which we will be adding information to
 * and advancing.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will add
 * the data items to the buffer until it reaches this pointer.  If it
 * goes to write to this address it will return -1.
 *
 * val - Unsigned integer value that we are packing into the buffer.
 */
extern
int	addbuf_put_uint(void **buf_pp, const void *bounds_p,
			const unsigned int val);

/*
 * int addbuf_get_uint
 *
 * DESCRIPTION:
 *
 * Get from the buffer a packed decimal value, advancing the buffer
 * pointer so it is always at the end of the data.  The value must
 * have been stored with the addbuf_put_uint.
 *
 * WARNING: this does not check the size of unsigned int and makes no
 * attempt to protect against overflow.
 *
 * RETURNS:
 *
 * Success - Number of characters taken from the buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to our buffer which we will be unpacking the value
 * from and advancing.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will get
 * the data items from the buffer until it reaches this pointer.  If
 * it goes to access this pointer, it will return -1.
 *
 * val_p - Pointer to an unsigned integer which will be assigned the
 * unpacked value.
 */
extern
int	addbuf_get_uint(void **buf_pp, const void *bounds_p,
			unsigned int *val_p);

/*
 * int addbuf_put_ulong
 *
 * DESCRIPTION:
 *
 * Add to a buffer a packed decimal value, advancing the buffer pointer
 * so it is always at the end of the data.  It will pack the decimal
 * into as few bytes as possible so addbuf_get_ulong must be used on
 * the other side to unpack.  This routine works up to values in
 * excess of (2^29-1).
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to our buffer which we will be adding information to
 * and advancing.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will add
 * the data items to the buffer until it reaches this pointer.  If it
 * goes to write to this address it will return -1.
 *
 * val - Unsigned long integer value that we are packing into the buffer.
 */
extern
int	addbuf_put_ulong(void **buf_pp, const void *bounds_p,
			 const unsigned long val);

/*
 * int addbuf_get_ulong
 *
 * DESCRIPTION:
 *
 * Get from the buffer a packed decimal value, advancing the buffer
 * pointer so it is always at the end of the data.  The value must
 * have been stored with the addbuf_put_ulong.
 *
 * WARNING: this does not check the size of unsigned long and makes no
 * attempt to protect against overflow.
 *
 * RETURNS:
 *
 * Success - Number of characters taken from the buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to our buffer which we will be unpacking the value
 * from and advancing.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will get
 * the data items from the buffer until it reaches this pointer.  If
 * it goes to access this pointer, it will return -1.
 *
 * val_p - Pointer to an unsigned long integer which will be assigned the
 * unpacked value.
 */
extern
int	addbuf_get_ulong(void **buf_pp, const void *bounds_p,
			 unsigned long *val_p);

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#endif /* ! __ADDBUF_H__ */
